''' Define the interface for a heuristic in order to implement any manner of heuristic
for a combinatorial problem that can be abstracted into a state graph. In this case, a
phylogenetic tree space. '''

# Date:   Apr 1 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

import os
from shutil import rmtree as removeFolder
from scoring import getLogLikelihoodForTopology as ll
from executable import raxml
from tree import tree

# Constants

SMGR_MIN_NUM_LIKELIHOOD = 32

# Base Heuristic Superclass/Interface

class heuristic(object):

    ''' A base interface for a heuristic that explores a state graph. '''
    
    def __init__(self,G=None,start=None):
        self.landscape = G      # State Graph
        self.start     = start  # Start State\
    def explore(self): pass
    def getStateGraph(self): return self.landscape
    def getStartState(self): return self.start

class phylogeneticLinearHeuristic(heuristic):
    
    ''' A base class for a heuristic that works on a phylogenetic landscape
    and only possesses a single path (of search). '''
    
    path, bestTree = list(), None
    def __init__(self,ls,startTree):
        super(phylogeneticLinearHeuristic,self
              ).__init__(ls,startTree)
    def getPath(self):     return self.path
    def getBestTree(self): return self.bestTree
    
# Heuristics

class parsimonyGreedy(phylogeneticLinearHeuristic):
    
    ''' Greedy (hill-climbing) landscape exploration by comparsion of parsimony. '''
    
    def __init__(self,ls,startTree):
        super(parsimonyGreedy,self).__init__(ls,startTree)
        
    def explore(self):
    
        ''' Perform greedy search of the landscape using
        a method of greed via parsimonious criterion. '''
        
        # Get starting tree, landscape.
        landscape = self.landscape
        ali       = landscape.alignment
        starting  = self.start
        cursor    = starting # Current tree.
        
        while (True):

            # Add this tree to the path.
            self.path.append(cursor)

            # Explore the current tree and only score via parsimony.
            landscape.exploreTree(cursor['tree'].name)
            
            # Rank by parsimony.
            nodes = landscape.graph.neighbors(cursor['tree'].name)
            nodes = [landscape.getNode(x) for x in nodes]
            nodes = [it for it in nodes if not it in self.path]
            nodes = sorted(nodes,key=lambda d: d['tree'].score[1])
            
            # Get best parsimony tree.
            best  = nodes[0]
            
            # Get parsimonies.
            bPars = best['tree'].score[1]
            cPars = cursor['tree'].score[1]
                 
            # Set new current tree if greater in parsimony.
            if (cPars > bPars): cursor = best
            else:
                self.bestTree = cursor
                break

class likelihoodGreedy(phylogeneticLinearHeuristic):
    
    ''' Greedy (hill-climbing) landscape exploration by comparsion of likelihood. '''
    
    def __init__(self,ls,startTree):
        super(likelihoodGreedy,self).__init__(ls,startTree)
        
    def explore(self):
    
        ''' Perform greedy search of the landscape using
        a method of greed via likelihood. '''
        
        # Get starting tree, landscape.
        landscape = self.landscape
        ali       = landscape.alignment
        starting  = self.start
        cursor    = starting # Current tree.
        
        while (True):

            # Add this tree to the path.
            self.path.append(cursor)

            # Explore the current tree and only score via parsimony.
            landscape.exploreTree(cursor['tree'].name)
            
            # Rank by likelihood.
            nodes = landscape.getNeighborsFor(cursor['tree'].name)
            nodes = [landscape.getNode(x) for x in nodes]
            nodes = [it for it in nodes if not it in self.path]
            map(lambda d: landscape.getVertex(d).scoreLikelihood(),nodes)
            nodes = sorted(nodes,key=lambda d: d['tree'].score[0])
            
            # Get best likelihood tree.
            best  = nodes[-1]
            
            # Get likelihoods.
            bLikl = best['tree'].score[0]
            cLikl = cursor['tree'].score[0]
                 
            # Set new current tree if greater in likelihood.
            if (cLikl > bLikl): cursor = best
            else:
                self.bestTree = cursor
                break

class smoothGreedy(phylogeneticLinearHeuristic):
    
    ''' Parsimony-driven greedy landscape exploration 
    by comparsion of likelihoods. '''
    
    def __init__(self,ls,startTree):
        super(smoothGreedy,self).__init__(ls,startTree)
        
    def explore(self):
    
        ''' Perform greedy search of the landscape using
        a method of greed via parsimonious criterion and then
        performing final smoothing via likelihood on top 10% of
        1-SPR neighbors ranked on basis of parsimony. '''
        
        # Get starting tree, landscape.
        landscape = self.landscape
        ali       = landscape.alignment
        starting  = self.start
        cursor    = starting # Current tree.
        
        while (True):

            # Add this tree to the path.
            self.path.append(cursor)

            # Explore the current tree and only score via parsimony.
            landscape.exploreTree(cursor['tree'].name)
            
            # Rank by parsimony.
            nodes = landscape.graph.neighbors(cursor['tree'].name)
            nodes = [landscape.getNode(x) for x in nodes]
            nodes = [it for it in nodes if not it in self.path]
            nodes = sorted(nodes,key=lambda d: d['tree'].score[1])
            
            # Get best parsimony tree.
            best  = nodes[0]
            
            # Get parsimonies.
            bPars = best['tree'].score[1]
            cPars = cursor['tree'].score[1]
            
            # Get likelihood of current tree if necessary.
            cL = cursor['tree'].score[0]
            if (cL == None):
                cL = ll(cursor['tree'].toTopology(),ali)
                cursor['tree'].score = (cL,cPars)            
            
            # Set new current tree if greater in parsimony.
            if (cPars > bPars): cursor = best
            else: # Smoothing via likelihood.
                # Check likelihood of some number of neighbors.
                numtodo = max(SMGR_MIN_NUM_LIKELIHOOD,len(nodes)/10)
                numtodo = min(numtodo,len(nodes))
                for node in nodes[:numtodo]:                
                    # Get log-likelihood of this neighbor.
                    if (node['tree'].score[0] == None):
                        ml = ll(node['tree'].toTopology(),ali)
                        node['tree'].score = (ml,node['tree'].score[1])
                    else: ml = node['tree'].score[0]
                    if (cL < ml):
                        # Short circuit.
                        cursor = node
                        continue
                # Rank by likelihood.
                visited = sorted(self.path,key=lambda d: d['tree'].score[0])
                nodes   = sorted(nodes,key=lambda d: d['tree'].score[0])
                # Get the best likelihood tree.
                vbest = visited[-1]
                best  = nodes[-1]
                if (vbest['tree'].score[0] > best['tree'].score[0]):
                    self.bestTree = vbest
                    break
                else: cursor = best
                
class RAxMLIdentify(phylogeneticLinearHeuristic):

    ''' RAxML-driven landscape evaluation
    of intermediate checkpoint trees output from the 
    RAxML executable. '''
    
    def __init__(self,ls,startNode,workdir='.rxml'):
        super(RAxMLIdentify,self).__init__(ls,startNode)
        self.executable = None
        self.workdir    = workdir        
        self.startFile  = os.path.join(workdir,'start')
        self.logFile    = os.path.join(workdir,'logfi')
        self.__setupWorkDir__()
        self.__setupExecutable__()
    
    def __setupWorkDir__(self):
        if not os.path.isdir(self.workdir):
            os.mkdir(self.workdir)
        o = open(self.startFile,'w')
        o.write(self.start['tree'].newick)
        o.close()
    
    def __setupExecutable__(self):
        ali = self.landscape.alignment
        phy = ali.getPhylip()
        r   = raxml(phy,'rx',interTrees=True,startingTree=self.startFile,
                    log=self.logFile,wdir=self.workdir)
        self.executable = r
    
    def __readLogFile__(self):
        checklogf = os.path.join(self.workdir,'RAxML_log.rx')
        if not os.path.isfile(checklogf):
            raise IOError('RAxML output logfile not found (%s).' % (checklogf))
        o = open(checklogf,'r')
        lines = o.readlines()
        iters = []
        for it in lines: iters.append(it.strip('\n').split(' '))
        o.close()
        return iters
        
    def __readIterTrees__(self,iters):
        trees = []
        for it in iters:
            chfi       = 'RAxML_checkpoint.rx.%s' % (it[-1])
            checkpoint = os.path.join(self.workdir,chfi)
            o          = open(checkpoint)
            tr         = tree(o.read().strip('\n'))
            o.close()
            trees.append(tr)
        return trees
    
    def explore(self):
        
        # Get the landscape.
        landscape = self.landscape
        
        try:
            
            # Run the RAxML executable, read output.
            self.executable.run()
            checkpoints = self.__readLogFile__()
            
            # Get all intermediate trees.
            trees = self.__readIterTrees__(checkpoints)
            
            # See if these trees are already in the landscape.
            for t in trees:
                find = landscape.findTreeTopology(t.newick)
                if not find:
                    t.origin = 'RAxML'
                    find = landscape.addTree(t)
                self.path.append(landscape.getNode(find))
                if t == trees[-1]: self.bestTree = find
                
            # Remove work directory.
            removeFolder(self.workdir)
            
        except Exception,e:
            removeFolder(self.workdir)
            raise IOError('RAxML I/O failure for reason [%s].' % (str(e)))
            
        


''' Container definition for (phylogenetic) bifurcating or multifurcating trees defined using Newick strings, collections of them, and for splits of these trees. '''

# Date:   Oct 20 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

# Imports

import newick, rearrangement, base
from math import factorial as fact
from numpy import median

# Function Defitions

numberRootedTrees   = lambda t: numberUnrootedTrees(t+1)
numberUnrootedTrees = lambda t: (fact(2*(t-1)-3))/((2**(t-3))*fact(t-3))

# Class Definitions for Trees

class tree(object): # TODO: Integrate with P4 Tree class (?).
    
    ''' Defines a single (phylogenetic) tree by newick string;
    can possess other metadata. '''
    
    def __init__(self,newi='',check=False):
        
        ''' If enabled, "check" will force the structure to reroot
        the given Newick string tree to a lowest-order leaf in order
        to ensure a consistent Newick string among any duplicate
        topologies. '''
        
        self.name   = ''
        self.score  = None
        self.origin = None
        self.newick = newi
        if (check): self._checkNewick(newi)        
        else:       self.setNewick(newi)
    
    # Getters, Mutators
    
    def getName(self):     return self.name
    def setName(self,n):   self.name = n
    def getScore(self):    return self.score
    def setScore(self,s):  self.score = s
    
    def getOrigin(self):   return self.origin
    def setOrigin(self,o):
        
        ''' Set the "origin" or specification of where this tree
        was acquired or constructed from; a string. '''
        
        self.origin = o    
        
    def getNewick(self):   return self.newick
    def toNewick(self):    return self.newick
    def setNewick(self,n):
        
        ''' Set Newick string to n; also reacquires corresponding
        "structure" or Newick string without branch lengths. '''
        
        self.newick = n
        self.struct = self._getStructure()   
        
    def getStructure(self):
        
        ''' Returns "structure", a Newick string without branch lengths. '''
        
        return self.struct 
    
    def getRerootedNoBranchLengthNewick(self): 
    
        ''' See getStructure(). '''
        
        return self.getStructure()
    
    def getSimpleNewick(self):

        ''' Return a Newick string with all taxa name replaced with
        successive integers. '''
    
        o = newick.parser(self.newick).parse()
        n = newick.getAllLeaves(o)
        for _ in xrange(1,len(n)+1): n[_].label = str(_)
        return str(o) + ';'        
    
    # Other Functionality
    
    def toTopology(self):
        
        ''' Return a rearrangement.topology instance for this tree to allow for 
        rearrangement of the actual structure of the tree. '''
        
        t = rearrangement.topology()
        t.fromNewick(self.newick)
        return t
    
    # Internals
        
    def __eq__(self,o): return (o.struct == self.struct)
    def __ne__(self,o): return not (self.__eq__(o))
    def __str__(self): return self.newick
    
    def _checkNewick(self,newi):
        
        ''' PRIVATE: Run the Newick string through a 
        parse pass and reroot to lowest-order leaf in
        order to ensure a consistent Newick string. '''
        
        newi        = newi.strip('\n').strip(';') + ';'   
        prsd        = newick.parser(newi).parse()
        self.newick = rearrangement.topology(prsd).toNewick()
        self.struct = self._getStructure(prsd)
    
    def _getStructure(self,prsd=None):
        
        ''' PRIVATE: Acquires a newick string without any
        defined branch lengths. '''
        
        if prsd: p = prsd
        else: p = newick.parser(self.newick).parse()    
        newick.removeBranchLengths(p)
        return str(p) + ';'  
    
class treeSet(base.Sized):
    
    ''' Represents an ordered, unorganized collection of trees 
    that do not necessarily comprise a combinatorial space. '''
    
    def __init__(self):
       
        self.trees = list()    
    
    def addTree(self,tr): 

        ''' Add a tree object to the collection. '''
    
        self.trees.append(tr)

    def addTreeByNewick(self,newick):
        
        ''' Add a tree to the structure by Newick string. '''
        
        t = tree(newick)
        self.addTree(t)    
        
    def removeTree(self,tr):
        
        ''' Remove a tree object from the collection if present. '''
        
        if tr in self.trees:
            self.trees.remove(tr)
    
    def indexOf(self,tr):
        
        ''' Acquire the index in this collection of a tree object. 
        Returns -1 if not found. '''
        
        if tr in self.trees: return self.trees.index(tr)
        else: return -1
    
    def __getitem__(self,i): return self.trees[i]
    def __len__(self): return len(self.trees)
    
    def toTreeFile(self,fout):
        
        ''' Output this landscape as a series of trees, separated by
        newlines, as a text file saved at the given path. '''        
        
        o = open(fout,'w')
        o.write(str(self))
        o.close()
        return fout
        
    def __str__(self):
        return '\n'.join([t.getNewick() for t in self.trees])
    
# Bipartition Class Definitions    
    
class bipartition(object):

    ''' A tree bipartition. Requires a tree topology. Using the term borrowed from nomenclature of a bipartite graph, a bipartition for a phylogenetic tree coincides with the definition of two disjoint sets U and V . A branch in a phylogenetic tree defines a single bipartition that divides the tree into two disjoint sets U and V . The set U comprises all of the children leaf of the subtree associated with that branch. The set V contains the rest of the leaves or taxa in the tree. '''

    def __init__(self,topol,bra=None):
        
        ''' Construct a bipartition from a branch in a topology. 
        
        :param topol: A topology.
        :type topol: :class: `rearrangement.topology`
        :param bra: An optional argument; can still acquire a bipartition from a string.
        :type bra: :class: `newick.branch`
        
        '''
        
        self.topology = topol
        self.branch   = bra
        self.btuple   = None # Tuple list representation.
        self.strrep   = None # String representation
        self.shortstr = ''   # Shorter string representation
        self.shortmap = None # Shorter string mapping of taxa to symbols
        self.reconfis = None # Possible reconfigurations as rearrangement objects.
        self._getStringRepresentation()
        self._getShortStringRepresentation()
        self._getBranchListRepresentation()
    
    def __hash__(self): return hash(self.shortstr)
    
    def __eq__(self,o):
        
        if (type(o) != bipartition): return False
        l,r   = self.strrep
        lo,ro = o.strrep
        return ((lo==l and ro==r) or (ro==l and lo==r))
    
    def __ne__(self,o): return not self.__eq__(o)
    
    def _getStringRepresentation(self):
        
        ''' PRIVATE: Obtain string representation of a branch in a topology. '''
        
        if (self.branch == None): return
        if (self.topology == None):
            raise ValueError('Topology does not exist or is equal to None.')
        
        z            = self.topology.getStrBipartitionFromBranch(self.branch)
        l,r          = sorted(z[0]),sorted(z[1])
        self.strrep  = (l,r)
        
    def _getShortStringRepresentation(self):
        
        ''' PRIVATE: Obtain short string representation of branch/bipartition. '''
        
        ind  = 65 # 'A'
        l,r  = self.strrep
        x    = max((l,r),key=lambda d:len(d))
        if x == l: y = r
        else: y = l
        l,r  = x,y
        o    = ''
        lbls = [x for x in l] + [x for x in r]
        lblm = {}
        lbls = sorted(lbls) # Sort them.
        for i in lbls:
            lblm[i] = chr(ind)
            ind += 1
        for i in l: o += lblm[i]
        o += ':'
        for i in r: o += lblm[i]
        self.shortstr = o
        self.shortmap = lblm
        
    def _getBranchListRepresentation(self):
        
        l = [self.branch] + newick.getAllBranches(self.branch)
        r = [branch for branch in self.topology.getBranches(
            ) if not branch in l]
        self.btuple = (l,r)       
        
    def _getBranchFromString(self):
        
        ''' PRIVATE: Obtain branch information from a string representation. '''
        
        if not self.strrep or type(self.strrep) != tuple:
            raise IOError('Could not read string representation of bipartition.')
        self.branch = self.topology.getBranchFromStrBipartition(self.strrep)
        
    def _makeSPRRearrangements(self):
        
        ''' PRIVATE: Perform possible rearrangements. '''
        
        self.reconfis = self.topology.allSPRForBranch(self.branch)

    def fromStringRepresentation(self,st):
        
        ''' Acquire all component elements from a string representation 
        of a bipartition. 
        
        :param st: A string representation from a :class:`.bipartition` object.
        
        '''
        
        self.strrep = st
        self._getBranchFromString()
        self._getShortStringRepresentation()
        
    def getBranch(self):
        
        ''' Get branch corresponding to this bipartition.
        
        :returns: :class:`newick.branch`
        
        '''
        
        return self.branch
    
    def getBranchIndex(self):
        
        ''' Return an index of the branch with respect to a post order traversal
        of the topology. '''
        
        nodes = base.treeStructure.postOrderTraversal(self.topology.getRoot())
        for i in xrange(len(nodes)):
            node = nodes[i]
            if node.parent == self.branch:
                return i
    
    def getStringRepresentation(self):
        
        ''' Get the string representation corresponding to this bipartition. '''
        
        return self.strrep
    
    def getShortStringRepresentation(self):
        
        ''' Get the shorter string representation corresponding to this bipartition. '''
        
        return self.shortstr
    
    def getShortStringMappings(self):
        
        ''' Get the mapping of symbols from taxa names for the shorter string representation. '''
        
        return self.shortmap
    
    def getBranchListRepresentation(self):
        
        ''' Get the tuple of lists of branches that represent this bipartition. '''
        
        return self.btuple
    
    def getSPRRearrangements(self):
        
        ''' Return the set of all scores related to this bipartition. '''
        
        if (self.reconfis == None): self._makeSPRRearrangements()
        return self.reconfis
    
    def getSPRScores(self,ls,node=None):
        
        ''' Given a landscape, return all possible scores, not actively performing
        scoring if not done. '''
        
        # Get starting information.
        scores = list() # Output scoreset.
        if not node:
            origin = self.topology # Starting topology.
            if not hasattr(origin,'orig'): ornewi = origin.toNewick()
            else: ornewi = origin.orig
            # Acquire node corresponding to this topology.
            start = ls.findTreeTopology(ornewi)
            if (start == None): return list() # Could not find.
        else: start = node 
            
        # Get neighbors.
        neighbors = ls.graph.neighbors(start)
        
        # Investigate all rearrangements.
        rearrangements = self.getSPRRearrangements()
        resultants = [x.toTree().struct for x in rearrangements]
        for neighbor in neighbors:
            node   = ls.getNode(neighbor)
            tree   = ls.getTree(neighbor)
            struct = tree.struct
            if struct in resultants:
                scores.append(tree.score[0])
        return scores
    
    def getMedianSPRScore(self,ls,node=None):
        
        ''' Given a landscape, return the median SPR score. '''
        
        li = [x for x in self.getSPRScores(ls,node) if x]
        if len(li) > 0: return median(li)
        else: return None
        
    def getBestSPRScore(self,ls,node=None):
        
        ''' Given a landscape, return the best SPR score. '''
        
        li = [x for x in self.getSPRScores(ls,node) if x]
        if len(li) > 0: return max(li)
        else: return None
